import os.path
from multiprocessing import Pool
from typing import Dict, Optional, Tuple
from geohash import encode, _base32
from .index import Index, IndexNotFound, PREBUILT_BASE_URL, PrebuiltIndexName
from .misc import generate_possible_hashes


class ShardedIndex:
    precision: int
    _split: int
    _idx: Dict[str, Index]

    def __init__(self, precision: int, *, shard_split: int = 1):
        self.precision = precision
        self._split = shard_split
        self._idx = {}

    def split_geohash(self, gh: str) -> Tuple[str, str]:
        return gh[:self._split], gh[self._split:]

    def add(self, gh: str, code: str):
        split, rest = self.split_geohash(gh)
        if split not in self._idx:
            self._idx[split] = Index(self.precision)
        self._idx[split].add(gh, code)

    def find_by_hash(self, gh: str) -> Optional[str]:
        split, rest = self.split_geohash(gh)
        if split not in self._idx:
            return None
        return self._idx[split].find_by_hash(gh)

    def find_by_coord(self, lat: float, lon: float) -> Optional[str]:
        gh = encode(lat, lon)
        return self.find_by_hash(gh)

    @staticmethod
    def filename(prefix: str, shard: str) -> str:
        return prefix + "_" + shard + ".idx"

    def dump(self, prefix: str, *, mkdir: bool = False):
        for shard, idx in self._idx.items():
            filename = self.filename(prefix, shard)
            idx.dump(filename, mkdir=mkdir)

    @staticmethod
    def _shard_loader(args: Tuple[str, str]) -> Optional[Tuple[str, Index]]:
        shard, filename = args
        try:
            idx = Index.load(filename)
            return shard, idx
        except IndexNotFound:
            return None

    @classmethod
    def load(cls, prefix: str, *, shard_split: int = 1, num_threads: int = 10):
        args = [(shard, cls.filename(prefix, shard)) for shard in generate_possible_hashes(shard_split)]
        with Pool(num_threads) as pool:
            results = filter(lambda v: v, pool.map(cls._shard_loader, args))

        indexes: Dict[str, Index] = dict(results)
        first = list(indexes.values())[0]
        precision = first.precision
        inst = cls(precision, shard_split=shard_split)
        inst._idx = indexes

        return inst

    @classmethod
    def load_prebuilt(cls, index_name: PrebuiltIndexName, precision: int):
        path = os.path.join(PREBUILT_BASE_URL, index_name, "sharded", str(precision), f"{index_name}{precision}")
        return cls.load(path, shard_split=1)



import json
import shapely.geometry
from geohash import _base32
from collections import deque
from typing import Optional, Generator
from .misc import geohash_poly, split_geohash


def load_country(code: str) -> Optional[shapely.geometry.base.BaseGeometry]:
    with open("countries.geojson") as f:
        data = json.load(f)

    for item in data["features"]:
        if item["id"] == code or item["properties"]["name"] == code:
            return shapely.geometry.shape(item["geometry"])
    return None


def collect(shape: shapely.geometry.base.BaseGeometry, *, max_precision: int = 5) -> Generator[str, None, None]:
    queue = deque(list(_base32))

    while queue:
        gh = queue.pop()
        poly = geohash_poly(gh)
        if poly.intersects(shape):
            if shape.contains(poly):
                yield gh
            elif len(gh) < max_precision:
                for sub_gh in split_geohash(gh):
                    queue.append(sub_gh)



from typing import Generator
from shapely import Polygon
from geohash import bbox, _base32


def geohash_poly(gh: str) -> Polygon:
    box = bbox(gh)
    coords = (
        (box["w"], box["n"]),
        (box["e"], box["n"]),
        (box["e"], box["s"]),
        (box["w"], box["s"]),
    )
    return Polygon(coords)


def split_geohash(gh: str) -> Generator[str, None, None]:
    for char in _base32:
        yield gh + char


def generate_possible_hashes(length: int) -> Generator[str, None, None]:
    max_value = len(_base32)
    digits = [0] * length
    while True:
        gh = "".join([_base32[i] for i in digits])
        yield gh
        upd = 0
        while upd < length:
            digits[upd] += 1
            if digits[upd] >= max_value:
                digits[upd] = 0
                upd += 1
            else:
                break
        if upd == length:
            return



import os.path
import requests
from typing import Optional, Self, Literal
from io import StringIO
from geohash import encode
from collections import defaultdict

PREBUILT_BASE_URL = "https://github.com/viert/worldgeo/raw/main/prebuilt"
PrebuiltIndexName = Literal["world"]


class IndexLoadError(Exception):
    pass


class IndexNotFound(Exception):
    pass


class SourceLoadError(Exception):
    pass


class InvalidSource(Exception):
    pass


class Index:
    precision: int
    _src: Optional[str]
    _index: dict[str, str]

    def __init__(self, precision: int):
        self.precision = precision
        self._index = {}
        self._src = None

    def add(self, gh: str, code: str):
        self._index[gh] = code

    def find_by_hash(self, gh: str) -> Optional[str]:
        for i in range(self.precision, 0, -1):
            key = gh[:i]
            if key in self._index:
                return self._index[key]
        return None

    def find_by_coord(self, lat: float, lon: float) -> Optional[str]:
        gh = encode(lat, lon)
        return self.find_by_hash(gh)

    def dump(self, filename: str, *, mkdir: bool = False):
        dmp = defaultdict(list)

        for gh, code in self._index.items():
            dmp[code].append(gh)

        if mkdir:
            dirname = os.path.dirname(filename)
            os.makedirs(dirname, exist_ok=True)

        with open(filename, "w") as f:
            f.write(f"P {self.precision}\n")
            for code, ghs in dmp.items():
                hashes = ":".join(ghs)
                f.write(f"G {code} {hashes}\n")

    @classmethod
    def load_prebuilt(cls, index_name: PrebuiltIndexName, precision: int) -> Self:
        path = os.path.join(PREBUILT_BASE_URL, index_name, f"world{precision}.idx")
        return cls.load(path)

    @classmethod
    def load(cls, path: str) -> Self:
        """
        Loads index from a file or HTTP(S) URL
        :param path: filename or HTTP url
        :return:
        """
        use_http = path.startswith("http://") or path.startswith("https://")
        if use_http:
            resp = requests.get(path)
            if resp.status_code == 404:
                raise IndexNotFound(f"index {path} not found")
            elif resp.status_code != 200:
                raise IndexLoadError(f"error loading index {path}: status code {resp.status_code}")
            f = StringIO(resp.text)
        else:
            if not os.path.isfile(path):
                raise IndexNotFound(f"index {path} not found")
            try:
                f = open(path)
            except EnvironmentError as e:
                raise IndexLoadError(f"error loading index {path}: {e}")

        precision = None
        index = {}

        for line in f:
            line = line.strip()
            tokens = line.split(" ", 2)

            match tokens[0]:
                case "P":
                    precision = int(tokens[1])
                case "G":
                    code = tokens[1]
                    ghs = tokens[2].split(":")
                    for gh in ghs:
                        index[gh] = code
        f.close()

        if precision is None:
            raise IndexLoadError("file is corrupted, precision cannot be found")

        inst = cls(precision)
        inst._src = path
        inst._index = index

        return inst

    def __repr__(self) -> str:
        source = f"\"{self._src}\"" if self._src else None
        return f"<Index source={source} precision={self.precision}>"



import json
import logging
import sys
import time
import requests
import os.path
from multiprocessing import Pool
from logging import getLogger, Logger, StreamHandler, Formatter
from typing import Optional, Dict, Any, Tuple, List
from queue import Queue
from shapely.geometry import shape
from shapely.geometry.base import BaseGeometry
from argparse import ArgumentParser
from worldgeo import Index, ShardedIndex
from worldgeo.collect import collect
from worldgeo.index import SourceLoadError, InvalidSource

DEFAULT_NUM_THREADS = 5
DEFAULT_GEOJSON_URL = "https://github.com/johan/world.geo.json/raw/master/countries.geo.json"

log: Logger


class Builder:

    src: str
    dest: str
    sharded: bool
    num_threads: int
    precision: int
    split_size: int
    mkdir: bool

    def __init__(self,
                 src: str,
                 dest: str,
                 precision: int,
                 *,
                 sharded: bool = False,
                 mkdir: bool = False,
                 num_threads: int = DEFAULT_NUM_THREADS,
                 split_size: int = 1):
        self.src = src
        self.dest = dest
        self.precision = precision
        self.sharded = sharded
        self.mkdir = mkdir
        self.num_threads = num_threads
        self.split_size = split_size

    def _load_source(self) -> Dict[str, Any]:
        use_http = self.src.startswith("http://") or self.src.startswith("https://")
        if use_http:
            log.debug(f"loading source from {self.src} via http")
            resp = requests.get(self.src)
            if resp.status_code == 404:
                raise SourceLoadError(f"source {self.src} not found")
            elif resp.status_code != 200:
                raise SourceLoadError(f"error loading index {self.src}: status code {resp.status_code}")
            return resp.json()
        else:
            if not os.path.isfile(self.src):
                raise SourceLoadError(f"index {self.src} not found")
            with open(self.src) as f:
                return json.load(f)

    def _collector(self, args: Tuple[str, BaseGeometry]) -> Tuple[str, List[str]]:
        code, boundaries = args
        ghs = list(collect(boundaries, max_precision=self.precision))
        return code, ghs

    def _build_index(self) -> ShardedIndex | Index:
        src = self._load_source()
        features = src.get("features")
        if features is None:
            raise InvalidSource("property \"features\" not found in source geojson")

        args = []
        count = 0
        failed_features = 0
        for feature in features:
            count += 1
            code = feature.get("id")
            if code is None:
                props = feature.get("properties")
                if props is None:
                    failed_features += 1
                    continue
                code = props.get("id")
                if code is None:
                    failed_features += 1
                    continue
            try:
                geom = shape(feature["geometry"])
            except:
                failed_features += 1
                continue
            args.append((code, geom))

        if failed_features > 0:
            if failed_features == count:
                raise InvalidSource("no valid features were found")
            else:
                log.warning("%d features have failed to process", failed_features)

        log.info("indexing %d features in %d threads", len(args), self.num_threads)

        with Pool(self.num_threads) as pool:
            results = pool.map(self._collector, args)

        if self.sharded:
            root = ShardedIndex(self.precision, shard_split=self.split_size)
        else:
            root = Index(self.precision)

        for res in results:
            code, ghs = res
            if ghs:
                log.debug("inserting %d geohashes for %s", len(ghs), code)
                for gh in ghs:
                    root.add(gh, code)

        return root

    def build(self) -> None:
        idx = self._build_index()
        idx.dump(self.dest, mkdir=self.mkdir)


def worker(wrk_id: int, in_queue: Queue, out_queue: Queue, max_precision: int):
    while True:
        item = in_queue.get()
        if item is None:
            break
        code, boundaries = item
        log.debug("[wrk %d] collecting geohash list for %s", wrk_id, code)
        t1 = time.time()
        for gh in collect(boundaries, max_precision=max_precision):
            out_queue.put((gh, code))
        t2 = time.time()
        log.debug("[wrk %d] collected geohash list for %s in %.3fs", wrk_id, code, t2-t1)
        in_queue.task_done()
    out_queue.put(None)
    log.debug("[wrk %d] finished", wrk_id)


def build_index(
    in_filename: Optional[str],
    out_filename: str,
    precision: int,
    *,
    sharded: bool,
    shard_split: int,
    country_filter: Optional[str] = None,
    num_threads: int = DEFAULT_NUM_THREADS,
):
    if sharded:
        root = ShardedIndex(precision, shard_split=shard_split)
    else:
        root = Index(precision)

    with open(in_filename) as f:
        data = json.load(f)

    in_queue = Queue()
    out_queue = Queue()

    if country_filter:
        country_filter = country_filter.lower()

    for item in data["features"]:
        country_name = item['properties']['name']
        country_code = item["id"]
        if country_filter and (
                country_name.lower() != country_filter and
                country_code.lower() != country_filter
        ):
            continue
        boundaries = shape(item["geometry"])
        in_queue.put((country_code, boundaries))

    for _ in range(num_threads):
        in_queue.put(None)

    wrk_alive = num_threads

    while True:
        item = out_queue.get()
        if item is None:
            wrk_alive -= 1
            if wrk_alive == 0:
                break
            continue
        gh, country = item
        root.add(gh, country)

    log.info("writing index to %s", out_filename)
    root.dump(out_filename)


def build():
    global log
    log = getLogger("indexer")
    log.setLevel(logging.DEBUG)
    for handler in log.handlers:
        log.removeHandler(handler)
    handler = StreamHandler(stream=sys.stdout)
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(Formatter("[%(asctime)s] %(levelname)s %(message)s"))
    log.addHandler(handler)

    parser = ArgumentParser()
    parser.add_argument("-i", "--input", default=None, help="geojson input filename")
    parser.add_argument("-o", "--output", required=True, help="index output filename "
                                                              "(or prefix for sharded indexes)")
    parser.add_argument("-p", "--precision", type=int, default=6, help="index maximum precision")
    parser.add_argument("-t", "--threads", type=int, default=5, help="number of threads")
    parser.add_argument("-s", "--sharded", default=False, action="store_true",
                        help="build sharded index")
    parser.add_argument("--split-size", type=int, default=1,
                        help="number of symbols to shard (sharded index only)")
    parser.add_argument("-m", "--mkdir", action="store_true", default=False,
                        help="automatically create output folders")
    args = parser.parse_args()

    builder = Builder(
        args.input or DEFAULT_GEOJSON_URL,
        args.output,
        args.precision,
        num_threads=args.threads,
        sharded=args.sharded,
        split_size=args.split_size,
        mkdir=args.mkdir,
    )
    builder.build()


if __name__ == "__main__":
    build()
